#!/usr/bin/env node
"use strict";
/**
 * typescript-build-linker
 * Link together TypeScript packages in a monorepo
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = __importStar(require("path"));
const trace_1 = require("@strong-roots-capital/trace");
const dependency_graph_1 = require("@typescript-tools/dependency-graph");
const io_ts_1 = require("@typescript-tools/io-ts");
const lerna_packages_1 = require("@typescript-tools/lerna-packages");
const lerna_utils_1 = require("@typescript-tools/lerna-utils");
const monorepo_root_1 = require("@typescript-tools/monorepo-root");
const stringify_json_1 = require("@typescript-tools/stringify-json");
const debug_1 = __importDefault(require("debug"));
const fast_deep_equal_1 = __importDefault(require("fast-deep-equal"));
const Apply_1 = require("fp-ts/Apply");
const A = __importStar(require("fp-ts/Array"));
const Console = __importStar(require("fp-ts/Console"));
const E = __importStar(require("fp-ts/Either"));
const Eq_1 = require("fp-ts/Eq");
const O = __importStar(require("fp-ts/Option"));
const R = __importStar(require("fp-ts/Record"));
const Semigroup_1 = require("fp-ts/Semigroup");
const S = __importStar(require("fp-ts/Set"));
const T = __importStar(require("fp-ts/Task"));
const TE = __importStar(require("fp-ts/TaskEither"));
const function_1 = require("fp-ts/function");
const string = __importStar(require("fp-ts/string"));
const get_relative_path_1 = __importDefault(require("get-relative-path"));
const t = __importStar(require("io-ts"));
const D = __importStar(require("io-ts-docopt"));
const io_ts_docopt_1 = require("io-ts-docopt");
const io_ts_types_1 = require("io-ts-types");
const PathReporter = __importStar(require("io-ts/lib/PathReporter"));
const ts_pattern_1 = require("ts-pattern");
const debug = {
    cmd: (0, debug_1.default)('link'),
};
// log to stdout instead of the default stderr
debug.cmd.log = console.log.bind(console);
const docstring = `
Usage:
    typescript-build-linker [<repository>]

Options:
    <repository>    Path to monorepo root, defaults to cwd
`;
const CommandLineOptions = (0, io_ts_docopt_1.withEncode)(t.type({
    '<repository>': (0, io_ts_types_1.withFallback)(t.string, process.cwd()),
}), (a) => ({
    repository: a['<repository>'],
}));
const monorepoRoot = (0, function_1.flow)(monorepo_root_1.monorepoRoot, TE.fromEither);
const decodeDocopt = (0, function_1.flow)(D.decodeDocopt, E.mapLeft((0, function_1.flow)((errors) => PathReporter.failure(errors).join('\n'), (error) => ({ type: 'docopt decode', error }))), TE.fromEither);
const readFile = (filename) => (0, function_1.pipe)((0, lerna_utils_1.readFile)(filename), TE.mapLeft((error) => ({ type: 'unable to read file', filename, error })));
const decodeFile = (codec) => (filename) => (contents) => (0, function_1.pipe)((0, io_ts_1.StringifiedJSON)(codec).decode(contents), E.mapLeft((0, function_1.flow)((errors) => PathReporter.failure(errors).join('\n'), (error) => ({ type: 'unexpected file contents', filename, error }))), TE.fromEither);
const stringifyJSON = (onError) => (json) => (0, function_1.pipe)((0, stringify_json_1.stringifyJSON)(onError)(json), E.mapLeft((error) => ({ type: 'unable to stringify json', json, error })), TE.fromEither);
const writeFile = (filename) => (contents) => (0, function_1.pipe)(contents, (0, trace_1.trace)(debug.cmd, `Writing file ${filename}`), (0, lerna_utils_1.writeFile)(filename), TE.mapLeft((error) => ({ type: 'unable to write file', filename, error })));
const writeJson = (filename) => (value) => (0, function_1.pipe)(stringifyJSON(E.toError)(value), TE.chainW(writeFile(filename)));
const writeJsonIfModified = (filename, original, desired) => (0, function_1.pipe)(O.fromPredicate((value) => !(0, fast_deep_equal_1.default)(original, value))(desired), O.fold((0, function_1.constant)(TE.right(undefined)), (0, function_1.constant)(writeJson(filename)(desired))));
const mapToRecord = (map) => Array.from(map.entries()).reduce((acc, [key, value]) => Object.assign(acc, { [key]: value }), {});
const parentDirectory = (directory) => (0, ts_pattern_1.match)(directory)
    .with('.', (0, function_1.constant)(O.none))
    .otherwise(() => O.some(path.dirname(directory)));
const ancestors = (directory) => {
    const ancestors = [];
    let parent = parentDirectory(directory);
    while (O.isSome(parent)) {
        (0, function_1.pipe)(parent, O.map((parent_) => {
            ancestors.push([parent_, path.basename(directory)]);
            directory = parent_;
            parent = parentDirectory(parent_);
        }));
    }
    return ancestors;
};
const linkChildrenPackages = (root) => (0, function_1.pipe)((0, lerna_packages_1.lernaPackages)(root), 
// map parents to children
TE.map((packages) => R.fromFoldableMap(A.getMonoid(), A.array)(packages, (pkg) => [
    // map parent directories into relative paths from root
    root === path.dirname(pkg.location)
        ? // treat the case where package is a direct child of monorepo-root,
            // because relativePath returns '../<root>' most irregularly
            '.'
        : (0, get_relative_path_1.default)(root + '/', path.dirname(pkg.location)),
    [pkg],
])), 
// isolate the lerna package directory
TE.map(R.map(A.map((pkg) => path.basename(pkg.location)))), 
// create ancestor references
TE.map((packagesByDirectory) => R.fromFoldableMap(S.getUnionMonoid(Eq_1.eqString), A.array)((0, function_1.pipe)(Object.entries(packagesByDirectory), A.chain(([directory, packages]) => [
    ...ancestors(directory),
    ...packages.map((pkg) => [directory, pkg]),
])), ([directory, pkg]) => [directory, new Set([pkg])])), 
// map set of children packages to set of children project references
TE.map(R.map((packages) => (0, function_1.pipe)(Array.from(packages), A.sort(string.Ord), A.map((pkg) => ({ path: pkg }))))), 
// map to write instructions
TE.map(R.mapWithIndex((parentDirectory, references) => {
    const tsconfigFile = path.resolve(root, parentDirectory, 'tsconfig.json');
    return (0, function_1.pipe)(readFile(tsconfigFile), TE.orElse((0, function_1.constant)(TE.right('{}'))), TE.chain(decodeFile(io_ts_1.TsConfig)(tsconfigFile)), 
    // The `files: []` prevents an accidental invocation
    // of `tsc` without `-b` from trying to build the
    // entire directory as one compilation:
    // https://github.com/RyanCavanaugh/learn-a
    TE.chain((tsconfig) => writeJsonIfModified(path.resolve(root, tsconfigFile), tsconfig, Object.assign({ files: [] }, tsconfig, { references }))));
})), 
// write in parallel
TE.chain((0, Apply_1.sequenceS)(TE.taskEither)), 
// return void on success
TE.map(function_1.constVoid));
const linkPackageDependencies = (root) => (0, function_1.pipe)(
// create map of package name to location (absolute path)
(0, lerna_packages_1.lernaPackages)(root), TE.map((packages) => R.fromFoldableMap((0, Semigroup_1.getLastSemigroup)(), A.array)(packages, (pkg) => [
    pkg.name,
    pkg.location,
])), TE.chain((lernaPackages) => (0, function_1.pipe)((0, dependency_graph_1.dependencyGraph)({ root, recursive: false }), TE.map(mapToRecord), TE.map(R.reduceWithIndex({}, (packageName, acc, dependencies) => Object.assign(acc, {
    [lernaPackages[packageName]]: (0, function_1.pipe)(dependencies, A.map((0, function_1.flow)((directory) => directory.location, (location) => path.relative(lernaPackages[packageName], location))), A.sort(string.Ord), A.map((path) => ({ path }))),
}))))), TE.map(R.mapWithIndex((packageDirectory, references) => {
    const tsconfigFile = path.resolve(packageDirectory, 'tsconfig.json');
    return (0, function_1.pipe)(readFile(tsconfigFile), TE.chain(decodeFile(io_ts_1.TsConfig)(tsconfigFile)), TE.chain((tsconfig) => writeJsonIfModified(tsconfigFile, tsconfig, Object.assign({}, tsconfig, { references }))));
})), 
// write in parallel
TE.chain((0, Apply_1.sequenceS)(TE.taskEither)), 
// return void on success
TE.map(function_1.constVoid));
const exit = (code) => () => process.exit(code);
const main = (0, function_1.pipe)(decodeDocopt(CommandLineOptions, docstring), TE.chainW((options) => monorepoRoot(options.repository)), TE.chain((root) => (0, function_1.pipe)([linkChildrenPackages(root), linkPackageDependencies(root)], TE.sequenceArray)), TE.fold((0, function_1.flow)(T.fromIOK(Console.error), T.chainIOK(() => exit(1))), (0, function_1.constant)(T.of(undefined))));
main();
// Local Variables:
// mode: typescript
// End:
//# sourceMappingURL=index.js.map